#define com_maroontress_libcgi_mime_ContentType_IMPORT
#define com_maroontress_libcgi_mime_Field_IMPORT
#define com_maroontress_libcgi_mime_Parameter_IMPORT
#define std_lang_String_IMPORT
#define std_util_ArraySet_IMPORT

#include <string.h>
#include <stdlib.h>
#include <com/maroontress/libcgi/mime/ContentType.h>
#include <com/maroontress/libcgi/mime/Field.h>
#include <com/maroontress/libcgi/mime/Parameter.h>
#include <std/lang/String.h>
#include <std/util/ArraySet.h>

struct ContentType {
    struct String *type;
    union ArraySet set;
};

static char *
trim(char *str)
{
    char *top;

    while (*str && strchr("\r\t ", *str) != NULL)
        ++str;
    top = str;
    while (*str)
        ++str;
    while (top < str && strchr("\r\t ", *str) != NULL)
        *str-- = 0;
    return (top);
}

static char *
dequote(char *str)
{
    int len;

    if ((len = strlen(str)) < 2
	|| str[0] != '"'
	|| str[len - 1] != '"') {
        return str;
    }
    str[len - 1] = 0;
    return str + 1;
}

static char *
separate(char **str, const char *delim)
{
    char *top, *p, *q;

    top = *str;
    if (*top == 0)
        return NULL;
    for (p = top; *p != 0 && strchr(delim, *p) == NULL; ++p) {
        switch (*p) {
        case '"':
            if (*(p = Field_endOfQuotedString(p)) == 0)
                return NULL;
            break;
        case '(':
	    if (*(q = Field_endOfComment(p)) == 0)
		return NULL;
	    memset(p, ' ', q - p + 1);
            break;
        case '\\':
            if (*++p == 0)
                return NULL;
            break;
        }
    }
    *str = (*p == 0) ? p : p + 1;
    *p = 0;
    return top;
}

static int
addParameter(struct ContentType *ct, char *p)
{
    size_t len;
    char *q, *name, *value;
    struct Parameter *param;

    /*
      p: param [ CWS ] * [ '=' [ CWS ] * value ] [ CWS ] *
    */
    if (*(p = trim(p)) == 0)
	return 0;
    len = strlen(p);
    q = separate(&p, "=");
    value = (strlen(q) < len) ? dequote(trim(p)) : NULL;
    name = trim(q);
    if ((param = Parameter_create(name, value)) == NULL)
	return -1;
    if (ArraySet_add(ct->set, param) < 0) {
	Parameter_free(param);
	return -1;
    }
    return 0;
}

static struct ContentType *
createContentType(char *s)
{
    struct ContentType *ct;
    char *p;

    /*
      s: maintype '/' subtype [ ';' param '=' value ] *
    */
    if ((ct = (struct ContentType *)malloc(sizeof(*ct))) == NULL) {
        return NULL;
    }
    ct->type = NULL;
    ct->set.impl = NULL;

    /*
      p: maintype '/' subtype
    */
    if ((p = separate(&s, ";")) == NULL
	|| *(p = trim(p)) == 0
	|| (ct->type = String_intern(p)) == NULL
	|| (ct->set = ArraySet_create()).impl == NULL) {
	ContentType_free(ct);
	return NULL;
    }
    while ((p = separate(&s, ";")) != NULL && addParameter(ct, p) == 0)
	;
    if (p != NULL) {
	ContentType_free(ct);
	return NULL;
    }
    return ct;
}

struct ContentType *
ContentType_create(const char *s)
{
    struct ContentType *ct;
    char *p;

    if ((p = strdup(s)) == NULL)
	return NULL;
    ct = createContentType(p);
    free(p);
    return ct;
}

void
ContentType_free(struct ContentType *ct)
{
    unsigned int k, n;

    if (ct == NULL)
	return;
    n = ArraySet_getSize(ct->set);
    for (k = 0; k < n; ++k)
	Parameter_free((struct Parameter *)ArraySet_get(ct->set, k));
    ArraySet_free(ct->set);
    String_free(ct->type);
    free(ct);
}

int
ContentType_matches(struct ContentType *ct,
		    const char *mainType, const char *subType)
{
    char *type;
    int len;

    type = String_getData(ct->type);
    len = strlen(mainType);
    if (strncasecmp(type, mainType, len) != 0 || type[len] != '/') {
        return 0;
    }
    return (subType == NULL || strcasecmp(type + len + 1, subType) == 0);
}

char *
ContentType_getType(struct ContentType *ct)
{
    return String_getData(ct->type);
}

struct Parameter *
ContentType_getParameter(struct ContentType *ct, const char *name)
{
    unsigned int k, n;
    struct Parameter *param;
    struct String *paramName;

    n = ArraySet_getSize(ct->set);
    for (k = 0; k < n; ++k) {
	param = (struct Parameter *)ArraySet_get(ct->set, k);
	paramName = Parameter_getName(param);
	if (strcasecmp(String_getData(paramName), name) == 0) {
	    return param;
	}
    }
    return NULL;
}
