#define com_maroontress_libcgi_request_Map_IMPORT
#define std_lang_String_IMPORT

#include <stdlib.h>
#include <string.h>
#include <com/maroontress/libcgi/request/Map.h>
#include <std/lang/String.h>

struct Pair {
    struct String *key;
    struct String *value;
};

struct Map {
    struct Pair *data;
    size_t size;
    size_t capacity;
};

void
Map_free(struct Map *map)
{
    unsigned int k;

    if (map == NULL)
	return;
    for (k = 0; k < map->size; ++k) {
	String_free(map->data[k].key);
	String_free(map->data[k].value);
    }
    free(map->data);
    free(map);
}

struct Map *
Map_create(void)
{
    struct Map *map;

    if ((map = (struct Map *)malloc(sizeof(*map))) == NULL)
	return NULL;
    map->data = NULL;
    map->size = 0;
    map->capacity = 0;
    return map;
}

static char *
get(struct Map *map, struct String *key)
{
    unsigned int k;

    for (k = 0; k < map->size && map->data[k].key != key; ++k)
	;
    return (k == map->size) ? NULL : String_getData(map->data[k].value);
}

char *
Map_get(struct Map *map, const char *key)
{
    struct String *s;
    char *v;

    if ((s = String_intern(key)) == NULL)
	return NULL;
    v = get(map, s);
    String_free(s);
    return v;
}

static int
put(struct Map *map, struct Pair *p)
{
    static const size_t delta = 8;
    struct Map next;
    size_t size;

    if ((next.size = map->size + 1) > map->capacity) {
	next.capacity = next.size + delta;
	size = next.capacity * sizeof(p);
	if ((next.data = (struct Pair *)realloc(map->data, size)) == NULL) {
	    return -1;
	}
	map->capacity = next.capacity;
	map->data = next.data;
    }
    map->data[map->size] = *p;
    map->size = next.size;
    return 0;
}

int
Map_put(struct Map *map, const char *key, const char *value)
{
    struct Pair p;

    p.key = String_intern(key);
    p.value = String_intern(value);
    if (p.key == NULL || p.value == NULL || put(map, &p) < 0) {
	String_free(p.key);
	String_free(p.value);
	return -1;
    }
    return 0;
}
